package postgres

import (
	"context"
	"database/sql"
	"errors"
	"testing"
	"time"

	"gitea.michaelthomson.dev/mthomson/habits/internal/migrate"
	"gitea.michaelthomson.dev/mthomson/habits/internal/todo/repository"
	_ "github.com/jackc/pgx/v5/stdlib"
	"github.com/testcontainers/testcontainers-go"
	"github.com/testcontainers/testcontainers-go/modules/postgres"
	"github.com/testcontainers/testcontainers-go/wait"
)

type TestDatabase struct {
	Db        *sql.DB
	container testcontainers.Container
}

func NewTestDatabase(tb testing.TB) *TestDatabase {
	tb.Helper()
	ctx := context.Background()
	// create container
	postgresContainer, err := postgres.Run(ctx,
		"postgres:16-alpine",
		postgres.WithDatabase("todo"),
		postgres.WithUsername("todo"),
		postgres.WithPassword("password"),
		testcontainers.WithWaitStrategy(
			wait.ForLog("database system is ready to accept connections").
				WithOccurrence(2).
				WithStartupTimeout(5*time.Second)),
	)

	if err != nil {
		tb.Fatalf("Failed to create postgres container, %v", err)
	}

	connectionString, err := postgresContainer.ConnectionString(ctx)
	if err != nil {
		tb.Fatalf("Failed to get connection string: %v", err)
	}

	// create db pool
	db, err := sql.Open("pgx", connectionString)
	if err != nil {
		tb.Fatalf("Failed to open db pool: %v", err)
	}

	migrate.Migrate(db)

	return &TestDatabase{
		Db:        db,
		container: postgresContainer,
	}
}

func (tdb *TestDatabase) TearDown() {
	tdb.Db.Close()
	_ = tdb.container.Terminate(context.Background())
}

func TestCRUD(t *testing.T) {
	tdb := NewTestDatabase(t)
	defer tdb.TearDown()
	r := NewPostgresTodoRepository(tdb.Db)

	t.Run("creates new todo", func(t *testing.T) {
		want := repository.TodoRow{Id: 1, Name: "clean dishes", Done: false}
		newTodo := repository.TodoRow{Name: "clean dishes", Done: false}
		got, err := r.Create(newTodo)
		AssertNoError(t, err)
		AssertTodoRows(t, got, want)
	})

	t.Run("gets todo", func(t *testing.T) {
		want := repository.TodoRow{Id: 1, Name: "clean dishes", Done: false}
		got, err := r.GetById(1)
		AssertNoError(t, err)
		AssertTodoRows(t, got, want)
	})

	t.Run("updates todo", func(t *testing.T) {
		want := repository.TodoRow{Id: 1, Name: "clean dishes", Done: true}
		err := r.Update(want)
		AssertNoError(t, err)

		got, err := r.GetById(1)
		AssertNoError(t, err)
		AssertTodoRows(t, got, want)
	})

	t.Run("deletes todo", func(t *testing.T) {
		err := r.Delete(1)
		AssertNoError(t, err)

		want := repository.ErrNotFound
		_, got := r.GetById(1)

		AssertErrors(t, got, want)
	})
}

func AssertErrors(t testing.TB, got, want error) {
	t.Helper()
	if !errors.Is(got, want) {
		t.Errorf("got error: %v, want error: %v", want, got)
	}
}

func AssertNoError(t testing.TB, err error) {
	t.Helper()
	if err != nil {
		t.Errorf("expected no error, got %v", err)
	}
}

func AssertTodoRows(t testing.TB, got, want repository.TodoRow) {
	t.Helper()
	if !got.Equal(want) {
		t.Errorf("got %+v want %+v", got, want)
	}
}
